# #
#	FILE:	    RandomMap.py
#	AUTHOR:     Terkhen (based on Sansimap by sansi)
# 	Version:    1.0_mst
#	PURPOSE:    Generate a map by selecting a random map script.
#   DEPENDENCY: Needs MapScriptTools.py
#-----------------------------------------------------------------------------
# CHANGELOG
#-----------------------------------------------------------------------------
#
#	1.0_mst       Terkhen  08.Dic.2014  ( uses MapScriptTools )
#     Added, RandomMap will try to avoid launching exceptions when the chosen mapscript lacks a certain method.
#     Added, save/load map options.
#     Added, weighted randomization of scenarios.
#     Added, support for Earth3, Erebus, Fractured World, Inland Sea, Medium and Small, Planetfall, Sea Highlands and Tectonics.
#     Added, initial version.
#-----------------------------------------------------------------------------

from CvPythonExtensions import *
import CvUtil
import CvMapGeneratorUtil

# Universal constants
map = CyMap()

# MapScriptTools Interface by Temudjin.
import MapScriptTools as mst

# Text information about the map script.
def getVersion():
	"""	Returns the version of the map script """
	return "1.0_mst"

def getDescription():
	"""	Returns the description of the map script """
	return "TXT_KEY_MAP_SCRIPT_RANDOM_DESCR"

# Methods called while we are still choosing the map options prior to map generation.

################################################################
## Custom Map Option Interface by Temudjin                 START
################################################################
def setCustomOptions():
	""" Loads custom map options from a cfg file (when possible) and configures all of them in a single place. """
	global op	# { optionID: Name, OptionList, Default, RandomOpt }

	# Initialize options to the default values.
	lMapOptions = [1, 0, 1, 2, 2, 2, 2, 2, 2, 2, 0, 0]

	# Try to read map options from the cfgFile.
	mst.mapOptionsStorage.initialize(lMapOptions, 'RandomMap')
	lMapOptions = mst.mapOptionsStorage.readConfig()

	optionWeights = [ "TXT_KEY_MAP_SCRIPT_RANDOM_NO", "1", "2", "3", "4" ]

	op = {
	       0: ["TXT_KEY_MAP_SCRIPT_WORLD_WRAP", ["TXT_KEY_MAP_SCRIPT_WORLD_WRAP_FLAT", "TXT_KEY_MAP_SCRIPT_WORLD_WRAP_CYLINDRICAL", "TXT_KEY_MAP_SCRIPT_WORLD_WRAP_TUBULAR", "TXT_KEY_MAP_SCRIPT_WORLD_WRAP_TOROIDAL"], lMapOptions[0], True],
		   1: ["TXT_KEY_MAP_RESOURCES", ["TXT_KEY_MAP_RESOURCES_STANDARD", "TXT_KEY_MAP_RESOURCES_BALANCED"], lMapOptions[1], True],
		   2: ["TXT_KEY_MAP_SCRIPT_RANDOM_EARTH3", optionWeights, lMapOptions[2], True],
		   3: ["TXT_KEY_MAP_SCRIPT_RANDOM_EREBUS", optionWeights, lMapOptions[3], True],
		   4: ["TXT_KEY_MAP_SCRIPT_RANDOM_FRACTURED_WORLD", optionWeights, lMapOptions[4], True],
		   5: ["TXT_KEY_MAP_SCRIPT_RANDOM_INLAND_SEA", optionWeights, lMapOptions[5], True],
		   6: ["TXT_KEY_MAP_SCRIPT_RANDOM_MEDIUM_AND_SMALL", optionWeights, lMapOptions[6], True],
		   7: ["TXT_KEY_MAP_SCRIPT_RANDOM_PLANETFALL", optionWeights, lMapOptions[7], True],
		   8: ["TXT_KEY_MAP_SCRIPT_RANDOM_SEA_HIGHLANDS", optionWeights, lMapOptions[8], True],
		   9: ["TXT_KEY_MAP_SCRIPT_RANDOM_TECTONICS", optionWeights, lMapOptions[9], True],
	       10: ["TXT_KEY_MAP_COASTS", ["TXT_KEY_MAP_COASTS_STANDARD", "TXT_KEY_MAP_COASTS_EXPANDED"], lMapOptions[10], True],
	       "Hidden": 10
	     }
	if mst.bPfall:
	    op[10] = ["TXT_KEY_MAP_TEAM_START", ["TXT_KEY_MAP_TEAM_START_NEIGHBORS", "TXT_KEY_MAP_TEAM_START_SEPARATED", "TXT_KEY_MAP_TEAM_START_RANDOM"], lMapOptions[10], False]
	elif mst.bMars:
		op[11] = ["TXT_KEY_MAP_MARS_THEME", ["TXT_KEY_MAP_MARS_THEME_SANDS_OF_MARS", "TXT_KEY_MAP_MARS_THEME_TERRAFORMED_MARS"], lMapOptions[11], False]
	else:
	    op[11] = ["TXT_KEY_MAP_TEAM_START", ["TXT_KEY_MAP_TEAM_START_NEIGHBORS", "TXT_KEY_MAP_TEAM_START_SEPARATED", "TXT_KEY_MAP_TEAM_START_RANDOM"], lMapOptions[11], False]

	mst.printDict(op,"Random Map Options:")

def isAdvancedMap():
	""" This map should show up in simple mode """
	return 0

# first function to be called by the map building process
def getNumHiddenCustomMapOptions():
	""" Default is used for the last n custom-options in 'Play Now' mode. """
	setCustomOptions()						# Define Options
	return op["Hidden"]

def getNumCustomMapOptions():
	""" Number of different user-defined options for this map """
	return len( op ) - 1

def getCustomMapOptionName(argsList):
	""" Returns name of specified option """
	optionID = argsList[0]
	translated_text = unicode(CyTranslator().getText(op[optionID][0], ()))
	return translated_text

def getNumCustomMapOptionValues(argsList):
	"""	Number of different choices for a particular setting """
	optionID = argsList[0]
	return len( op[optionID][1] )

def getCustomMapOptionDescAt(argsList):
	""" Returns name of value of option at specified row """
	optionID = argsList[0]
	valueID = argsList[1]
	translated_text = unicode(CyTranslator().getText(op[optionID][1][valueID], ()))
	return translated_text

def getCustomMapOptionDefault(argsList):
	"""	Returns default value of specified option """
	optionID = argsList[0]
	return op[optionID][2]

def isRandomCustomMapOption(argsList):
	"""	Returns a flag indicating whether a random option should be provided """
	optionID = argsList[0]
	return op[optionID][3]

################################################################
## Interfaces by Temudjin                                    END
################################################################

def isClimateMap():
	"""	Indicates if this map script uses the Climate options. """
	# At this point of map generation we still don't know if the map script to be chosen randomly
	# will require this option, so we always request it.
	return True

def isSeaLevelMap():
	"""	Indicates if this map script uses the Sea Level options. """
	# At this point of map generation we still don't know if the map script to be chosen randomly
	# will require this option, so we always request it.
	return True

# Methods called before map generation is initialized.

def beforeInit():
	"""	Obtains a map chosen randomly from the weights given by the user and initializes it. """

	# Map script to be chosen randomly.
	global cms

	# Calculate weights for each map
	pMapScripts = []
	print "[RandomMap] Calculating map script weights..."
	iTotalWeight = 0
	if map.getCustomMapOption(2) != 0:
		pMapScripts.append((map.getCustomMapOption(2), 'EARTH3'))
		iTotalWeight = iTotalWeight + map.getCustomMapOption(2)
		print "[RandomMap] EARTH3 weight: %i (Total weight: %i)" % (map.getCustomMapOption(2), iTotalWeight)
	if map.getCustomMapOption(3) != 0:
		pMapScripts.append((map.getCustomMapOption(3), 'EREBUS'))
		iTotalWeight = iTotalWeight + map.getCustomMapOption(3)
		print "[RandomMap] EREBUS weight: %i (Total weight: %i)" % (map.getCustomMapOption(3), iTotalWeight)
	if map.getCustomMapOption(4) != 0:
		pMapScripts.append((map.getCustomMapOption(4), 'FRACTURED_WORLD'))
		iTotalWeight = iTotalWeight + map.getCustomMapOption(4)
		print "[RandomMap] FRACTURED_WORLD weight: %i (Total weight: %i)" % (map.getCustomMapOption(4), iTotalWeight)
	if map.getCustomMapOption(5) != 0:
		pMapScripts.append((map.getCustomMapOption(5), 'INLAND_SEA'))
		iTotalWeight = iTotalWeight + map.getCustomMapOption(5)
		print "[RandomMap] INLAND_SEA weight: %i (Total weight: %i)" % (map.getCustomMapOption(5), iTotalWeight)
	if map.getCustomMapOption(6) != 0:
		pMapScripts.append((map.getCustomMapOption(6), 'MEDIUM_AND_SMALL'))
		iTotalWeight = iTotalWeight + map.getCustomMapOption(6)
		print "[RandomMap] MEDIUM_AND_SMALL weight: %i (Total weight: %i)" % (map.getCustomMapOption(6), iTotalWeight)
	if map.getCustomMapOption(7) != 0:
		pMapScripts.append((map.getCustomMapOption(7), 'PLANETFALL'))
		iTotalWeight = iTotalWeight + map.getCustomMapOption(7)
		print "[RandomMap] PLANETFALL weight: %i (Total weight: %i)" % (map.getCustomMapOption(7), iTotalWeight)
	if map.getCustomMapOption(8) != 0:
		pMapScripts.append((map.getCustomMapOption(8), 'SEA_HIGHLANDS'))
		iTotalWeight = iTotalWeight + map.getCustomMapOption(8)
		print "[RandomMap] SEA_HIGHLANDS weight: %i (Total weight: %i)" % (map.getCustomMapOption(8), iTotalWeight)
	if map.getCustomMapOption(9) != 0:
		pMapScripts.append((map.getCustomMapOption(9), 'TECTONICS'))
		iTotalWeight = iTotalWeight + map.getCustomMapOption(9)
		print "[RandomMap] TECTONICS weight: %i (Total weight: %i)" % (map.getCustomMapOption(9), iTotalWeight)

	# If the user chose no map scripts, we will choose a random one between all supported map scripts.
	if len(pMapScripts) == 0:
		pMapScripts.append((1, 'EARTH3'))
		pMapScripts.append((2, 'EREBUS'))
		pMapScripts.append((2, 'FRACTURED_WORLD'))
		pMapScripts.append((2, 'INLAND_SEA'))
		pMapScripts.append((2, 'MEDIUM_AND_SMALL'))
		pMapScripts.append((2, 'PLANETFALL'))
		pMapScripts.append((2, 'SEA_HIGHLANDS'))
		pMapScripts.append((2, 'TECTONICS'))

	# Choose randomly between all supported mapscripts, taking into account their weights.
	iRand = CyGlobalContext().getGame().getMapRand().get(iTotalWeight, "RandomMap.beforeInit()")
	iCurrentWeight = 0
	for index in range( len(pMapScripts) ):
		iWeight, sMapScript = pMapScripts[index]
		iCurrentWeight = iCurrentWeight + iWeight
		print "[RandomMap] %s current weight: %i (Random value: %i)" % (sMapScript, iCurrentWeight, iRand)
		if iRand < iCurrentWeight:
			sChosenMapScript = sMapScript
			print "[RandomMap] %s has been chosen" % (sMapScript)
			break

	# Initialize the chosen map script.
	if sChosenMapScript == 'EARTH3':
		import Earth3_mst as cms
	if sChosenMapScript == 'EREBUS':
		import Erebus_mst as cms
	if sChosenMapScript == 'FRACTURED_WORLD':
		import FracturedWorld_mst as cms
	if sChosenMapScript == 'INLAND_SEA':
		import Inland_Sea_mst as cms
	if sChosenMapScript == 'MEDIUM_AND_SMALL':
		import Medium_and_Small_mst as cms
	if sChosenMapScript == 'PLANETFALL':
		import PlanetFall_mst as cms
	if sChosenMapScript == 'SEA_HIGHLANDS':
		import Sea_Highlands_mst as cms
	if sChosenMapScript == 'TECTONICS':
		import Tectonics_mst as cms

	# Map options to be transferred to the chosen map script.
	mapOptionWorldShape = map.getCustomMapOption(0)
	mapOptionResources = map.getCustomMapOption(1)
	mapOptionCoastalWaters = mst.iif(mst.bPfall, None, map.getCustomMapOption(10))
	mapOptionTeamStart = mst.iif(mst.bMars, None, map.getCustomMapOption( mst.iif(mst.bPfall,10,11) ))
	mapOptionMarsTheme = mst.iif(mst.bMars, map.getCustomMapOption(11), None)

	# The chosen map script will initialize its map options with these values. All other map options will be set randomly.
	try:
		cmsMethod = getattr(cms, "randomMapBeforeInit")
	except AttributeError:
		print "[RandomMap] Fatal error: %s is not compatible with RandomMap" % (sChosenMapScript)
	else:
		cmsMethod(mapOptionWorldShape, mapOptionResources, mapOptionCoastalWaters, mapOptionTeamStart, mapOptionMarsTheme)

def getGridSize(argsList):
	"""	Returns the chosen grid size. """
	if (argsList[0] == -1): # (-1,) is passed to function on loads
		return []
	try:
		cmsMethod = getattr(cms, "method")
	except AttributeError:
		return []
	else:
		return cmsMethod(argsList)

def getTopLatitude():
	""" Returns the maximum latitude. This value must be in a [-90, 90] interval. """
	try:
		cmsMethod = getattr(cms, "getTopLatitude")
	except AttributeError:
		return 90
	else:
		return cmsMethod()

def getBottomLatitude():
	""" Returns the minimum latitude. This value must be in a [-90, 90] interval. """
	try:
		cmsMethod = getattr(cms, "getBottomLatitude")
	except AttributeError:
		return -90
	else:
		return cmsMethod()

def getWrapX():
	""" Returns if the map should wrap in the X axis. """
	try:
		cmsMethod = getattr(cms, "getWrapX")
	except AttributeError:
		return False
	else:
		return cmsMethod()

def getWrapY():
	""" Returns if the map should wrap in the Y axis. """
	try:
		cmsMethod = getattr(cms, "getWrapY")
	except AttributeError:
		return False
	else:
		return cmsMethod()

# Methods called before map generation.

def beforeGeneration():
	""" Saves the map options used for this generation and calls beforeGeneration() in the chosen map script. """

	# Obtain the map options in use by RandomMap.
	lMapOptions = []
	for opt in range( getNumCustomMapOptions() ):
		iValue = 0 + map.getCustomMapOption( opt )
		lMapOptions.append( iValue )

	# Save used RandomMap map options.
	mst.mapOptionsStorage.writeConfig(lMapOptions)

	# Call beforeGeneration() in the chosen map script.
	try:
		cmsMethod = getattr(cms, "beforeGeneration")
	except AttributeError:
		return False
	else:
		return cmsMethod()

# Methods called during map generation.

def generatePlotTypes():
	""" FIRST stage in map generation. Calls generatePlotTypes() in the chosen map script. """
	try:
		cmsMethod = getattr(cms, "generatePlotTypes")
	except AttributeError:
		return CyPythonMgr().allowDefaultImpl()
	else:
		return cmsMethod()

def generateTerrainTypes():
	""" SECOND stage in map generation. Calls generateTerrainTypes() in the chosen map script. """
	try:
		cmsMethod = getattr(cms, "generateTerrainTypes")
	except AttributeError:
		return CyPythonMgr().allowDefaultImpl()
	else:
		return cmsMethod()

def addRivers():
	""" THIRD stage in map generation. Calls addRivers() in the chosen map script. """
	try:
		cmsMethod = getattr(cms, "addRivers")
	except AttributeError:
		CyPythonMgr().allowDefaultImpl()
	else:
		cmsMethod()

def addLakes():
	""" FOURTH stage in map generation. Calls addLakes() in the chosen map script. """
	try:
		cmsMethod = getattr(cms, "addLakes")
	except AttributeError:
		CyPythonMgr().allowDefaultImpl()
	else:
		cmsMethod()

def addFeatures():
	""" FIFTH stage in map generation. Calls addFeatures() in the chosen map script. """
	try:
		cmsMethod = getattr(cms, "addFeatures")
	except AttributeError:
		CyPythonMgr().allowDefaultImpl()
	else:
		cmsMethod()

def addBonuses():
	""" SIXTH stage in map generation. Calls addBonuses() in the chosen map script. """
	try:
		cmsMethod = getattr(cms, "addBonuses")
	except AttributeError:
		CyPythonMgr().allowDefaultImpl()
	else:
		cmsMethod()

def isBonusIgnoreLatitude():
	""" Defines if bonus placement should ignore map latitude. Used by the default implementation of addBonuses(), normalizeAddFoodBonuses() and normalizeAddExtras(). Calls isBonusIgnoreLatitude() in the chosen map script. """
	try:
		cmsMethod = getattr(cms, "isBonusIgnoreLatitude")
	except AttributeError:
		return False
	else:
		return cmsMethod()

def addGoodies():
	""" SEVENTH stage in map generation. Calls addGoodies() in the chosen map script. """
	try:
		cmsMethod = getattr(cms, "addGoodies")
	except AttributeError:
		CyPythonMgr().allowDefaultImpl()
	else:
		cmsMethod()

# Methods called during starting plot normalization.

def normalizeStartingPlotLocations():
	""" FIRST stage in the normalization of starting plots. Calls normalizeStartingPlotLocations() in the chosen map script. """
	try:
		cmsMethod = getattr(cms, "normalizeStartingPlotLocations")
	except AttributeError:
		CyPythonMgr().allowDefaultImpl()
	else:
		cmsMethod()

def normalizeAddRiver():
	""" SECOND stage in the normalization of starting plots. Calls normalizeAddRiver() in the chosen map script. """
	try:
		cmsMethod = getattr(cms, "normalizeAddRiver")
	except AttributeError:
		CyPythonMgr().allowDefaultImpl()
	else:
		cmsMethod()

def normalizeRemovePeaks():
	""" THIRD stage in the normalization of starting plots. Calls normalizeRemovePeaks() in the chosen map script. """
	try:
		cmsMethod = getattr(cms, "normalizeRemovePeaks")
	except AttributeError:
		return CyPythonMgr().allowDefaultImpl()
	else:
		return cmsMethod()

def normalizeAddLakes():
	""" FOURTH stage in the normalization of starting plots. Calls normalizeAddLakes() in the chosen map script. """
	try:
		cmsMethod = getattr(cms, "normalizeAddLakes")
	except AttributeError:
		CyPythonMgr().allowDefaultImpl()
	else:
		cmsMethod()

def normalizeRemoveBadFeatures():
	""" FIFTH stage in the normalization of starting plots. Calls normalizeRemoveBadFeatures() in the chosen map script. """
	try:
		cmsMethod = getattr(cms, "normalizeRemoveBadFeatures")
	except AttributeError:
		return CyPythonMgr().allowDefaultImpl()
	else:
		return cmsMethod()

def normalizeRemoveBadTerrain():
	""" SIXTH stage in the normalization of starting plots. Calls normalizeRemoveBadTerrain() in the chosen map script. """
	try:
		cmsMethod = getattr(cms, "normalizeRemoveBadTerrain")
	except AttributeError:
		CyPythonMgr().allowDefaultImpl()
	else:
		cmsMethod()

def normalizeAddFoodBonuses():
	""" SEVENTH stage in the normalization of starting plots. Calls normalizeAddFoodBonuses() in the chosen map script. """
	try:
		cmsMethod = getattr(cms, "normalizeAddFoodBonuses")
	except AttributeError:
		CyPythonMgr().allowDefaultImpl()
	else:
		cmsMethod()

def normalizeAddGoodTerrain():
	""" EIGHT stage in the normalization of starting plots. Calls normalizeAddGoodTerrain() in the chosen map script. """
	try:
		cmsMethod = getattr(cms, "normalizeAddGoodTerrain")
	except AttributeError:
		CyPythonMgr().allowDefaultImpl()
	else:
		cmsMethod()

def normalizeAddExtras():
	""" NINTH and last stage in the normalization of starting plots. Calls normalizeAddExtras() in the chosen map script. """
	try:
		cmsMethod = getattr(cms, "normalizeAddExtras")
	except AttributeError:
		CyPythonMgr().allowDefaultImpl()
	else:
		cmsMethod()

def minStartingDistanceModifier():
	""" FIRST stage in the selection of starting plots. Calls minStartingDistanceModifier() in the chosen map script. """
	try:
		cmsMethod = getattr(cms, "minStartingDistanceModifier")
	except AttributeError:
			if mst.bPfall: return -25
			if mst.bMars: return -15
			return 0
	else:
		return cmsMethod()

##########
